@@ -32,24 +32,23 @@
  * files in the program, then also delete it here.
  */
 /* copyright --> */
-#ifndef _D_EPOLL_EVENT_POLL_H_
-#define _D_EPOLL_EVENT_POLL_H_
+#ifndef D_EPOLL_EVENT_POLL_H
+#define D_EPOLL_EVENT_POLL_H
 
 #include "EventPoll.h"
 
 # include <sys/epoll.h>
 
-#include <deque>
+#include <set>
 
 #include "Event.h"
+#include "a2functional.h"
 #ifdef ENABLE_ASYNC_DNS
 # include "AsyncNameResolver.h"
 #endif // ENABLE_ASYNC_DNS
 
 namespace aria2 {
 
-class Logger;
-
 class EpollEventPoll : public EventPoll {
 private:
   class KSocketEntry;
@@ -71,21 +70,24 @@ private:
   friend int accumulateEvent(int events, const KEvent& event);
 
 private:
-  std::deque<SharedHandle<KSocketEntry> > _socketEntries;
+  typedef std::set<SharedHandle<KSocketEntry>,
+                   DerefLess<SharedHandle<KSocketEntry> > > KSocketEntrySet;
+  KSocketEntrySet socketEntries_;
 #ifdef ENABLE_ASYNC_DNS
-  std::deque<SharedHandle<KAsyncNameResolverEntry> > _nameResolverEntries;
+  typedef std::set<SharedHandle<KAsyncNameResolverEntry>,
+                   DerefLess<SharedHandle<KAsyncNameResolverEntry> > >
+  KAsyncNameResolverEntrySet;
+  KAsyncNameResolverEntrySet nameResolverEntries_;
 #endif // ENABLE_ASYNC_DNS
 
-  int _epfd;
+  int epfd_;
 
-  size_t _epEventsSize;
+  size_t epEventsSize_;
 
-  struct epoll_event* _epEvents;
+  struct epoll_event* epEvents_;
 
   static const size_t EPOLL_EVENTS_MAX = 1024;
 
-  Logger* _logger;
-
   bool addEvents(sock_t socket, const KEvent& event);
 
   bool deleteEvents(sock_t socket, const KEvent& event);
@@ -126,4 +128,4 @@ public:
 
 } // namespace aria2
 
-#endif // _D_EPOLL_EVENT_POLL_H_
+#endif // D_EPOLL_EVENT_POLL_H
