@@ -44,7 +44,7 @@
 #include "Option.h"
 #include "Logger.h"
 #include "LogFactory.h"
-#include "Socket.h"
+#include "SocketCore.h"
 #include "message.h"
 #include "prefs.h"
 #include "A2STR.h"
@@ -82,6 +82,7 @@ Command* HttpInitiateConnectionCommand::createNextCommand
                       getCuid(), addr.c_str(), port));
       createSocket();
       getSocket()->establishConnection(addr, port);
+      getSocket()->setTcpNodelay(true);
 
       getRequest()->setConnectedAddrInfo(hostname, addr, port);
       if(proxyMethod == V_TUNNEL) {
@@ -139,6 +140,8 @@ Command* HttpInitiateConnectionCommand::createNextCommand
                       getCuid(), addr.c_str(), port));
       createSocket();
       getSocket()->establishConnection(addr, port);
+      getSocket()->setTcpNodelay(true);
+
       getRequest()->setConnectedAddrInfo(hostname, addr, port);
     } else {
       setSocket(pooledSocket);
