@@ -32,8 +32,8 @@
  * files in the program, then also delete it here.
  */
 /* copyright --> */
-#ifndef _D_OPTION_PARSER_H_
-#define _D_OPTION_PARSER_H_
+#ifndef D_OPTION_PARSER_H
+#define D_OPTION_PARSER_H
 
 #include "common.h"
 
@@ -47,65 +47,58 @@ namespace aria2 {
 
 class Option;
 class OptionHandler;
+struct Pref;
 
 class OptionParser {
 private:
-  int idCounter_;
-
-  // optionHandlers_ is sorted by OptionHandler::getName() in
-  // ascending order.
-  std::vector<SharedHandle<OptionHandler> > optionHandlers_;
-
-  SharedHandle<OptionHandler>
-  getOptionHandlerByName(const std::string& optName);
-
+  std::vector<OptionHandler*> handlers_;
+  // Index of handler in handlers_ for option who has short option name.
+  std::vector<size_t> shortOpts_;
   static SharedHandle<OptionParser> optionParser_;
 public:
   OptionParser();
-
-  ~OptionParser() {}
+  ~OptionParser();
 
   // Parses options in argv and writes option name and value to out in
   // NAME=VALUE format. Non-option strings are stored in nonopts.
   // Throws Exception when an unrecognized option is found.
   void parseArg(std::ostream& out, std::vector<std::string>& nonopts,
-                int argc, char* const argv[]);
+                int argc, char* argv[]) const;
 
-  void parse(Option& option, std::istream& ios);
+  void parse(Option& option, std::istream& ios) const;
 
   void parseDefaultValues(Option& option) const;
 
   void setOptionHandlers
-  (const std::vector<SharedHandle<OptionHandler> >& optionHandlers);
+  (const std::vector<OptionHandler*>& handlers);
 
-  void addOptionHandler(const SharedHandle<OptionHandler>& optionHandler);
+  void addOptionHandler(OptionHandler* handler);
 
   // Hidden options are not returned.
-  std::vector<SharedHandle<OptionHandler> >
-  findByTag(const std::string& tag) const;
+  std::vector<const OptionHandler*> findByTag(uint32_t tag) const;
 
   // Hidden options are not returned.
-  std::vector<SharedHandle<OptionHandler> >
+  std::vector<const OptionHandler*>
   findByNameSubstring(const std::string& substring) const;
 
   // Hidden options are not returned.
-  std::vector<SharedHandle<OptionHandler> > findAll() const;
+  std::vector<const OptionHandler*> findAll() const;
 
   // Hidden options are not returned.
-  SharedHandle<OptionHandler>
-  findByName(const std::string& name) const;
+  const OptionHandler* find(const Pref* pref) const;
 
   // Hidden options are not returned.
-  SharedHandle<OptionHandler> findByID(int id) const;
+  const OptionHandler* findById(size_t id) const;
 
   // Hidden options are not returned.
-  SharedHandle<OptionHandler> findByShortName(char shortName) const;
+  const OptionHandler* findByShortName(char shortName) const;
 
   static const SharedHandle<OptionParser>& getInstance();
+  // Deletes statically allocated instace. Call this at the end of the
+  // program.
+  static void deleteInstance();
 };
 
-typedef SharedHandle<OptionParser> OptionParserHandle;
-
 } // namespace aria2
 
-#endif // _D_OPTION_PARSER_H_
+#endif // D_OPTION_PARSER_H
