@@ -42,6 +42,11 @@
 #include <cerrno>
 #include <cstring>
 
+#ifdef HAVE_OPENSSL
+# include <openssl/x509.h>
+# include <openssl/x509v3.h>
+#endif // HAVE_OPENSSL
+
 #ifdef HAVE_LIBGNUTLS
 # include <gnutls/x509.h>
 #endif // HAVE_LIBGNUTLS
@@ -89,7 +94,7 @@ namespace aria2 {
 #ifdef __MINGW32__
 # define CLOSE(X) ::closesocket(X)
 #else
-# define CLOSE(X) while(close(X) == -1 && errno == EINTR)
+# define CLOSE(X) close(X)
 #endif // __MINGW32__
 
 namespace {
@@ -104,7 +109,7 @@ std::string errorMsg(int errNum)
                     FORMAT_MESSAGE_IGNORE_INSERTS,
                     NULL,
                     errNum,
-                    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
+                    MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
                     (LPTSTR) &buf,
                     sizeof(buf),
                     NULL
@@ -116,17 +121,36 @@ std::string errorMsg(int errNum)
 }
 } // namespace
 
+namespace {
+enum TlsState {
+  // TLS object is not initialized.
+  A2_TLS_NONE = 0,
+  // TLS object is now handshaking.
+  A2_TLS_HANDSHAKING = 2,
+  // TLS object is now connected.
+  A2_TLS_CONNECTED = 3
+};
+} // namespace
+
 int SocketCore::protocolFamily_ = AF_UNSPEC;
 
-std::vector<std::pair<struct sockaddr_storage, socklen_t> >
+std::vector<std::pair<sockaddr_union, socklen_t> >
 SocketCore::bindAddrs_;
 
 #ifdef ENABLE_SSL
-SharedHandle<TLSContext> SocketCore::tlsContext_;
+SharedHandle<TLSContext> SocketCore::clTlsContext_;
+SharedHandle<TLSContext> SocketCore::svTlsContext_;
 
-void SocketCore::setTLSContext(const SharedHandle<TLSContext>& tlsContext)
+void SocketCore::setClientTLSContext
+(const SharedHandle<TLSContext>& tlsContext)
 {
-  tlsContext_ = tlsContext;
+  clTlsContext_ = tlsContext;
+}
+
+void SocketCore::setServerTLSContext
+(const SharedHandle<TLSContext>& tlsContext)
+{
+  svTlsContext_ = tlsContext;
 }
 #endif // ENABLE_SSL
 
@@ -147,7 +171,7 @@ SocketCore::SocketCore(sock_t sockfd, int sockType)
 void SocketCore::init()
 {
   blocking_ = true;
-  secure_ = 0;
+  secure_ = A2_TLS_NONE;
 
   wantRead_ = false;
   wantWrite_ = false;
@@ -173,8 +197,7 @@ void SocketCore::create(int family, int protocol)
   errNum = SOCKET_ERRNO;
   if(fd == (sock_t) -1) {
     throw DL_ABORT_EX
-      (fmt("Failed to create socket. Cause:%s",
-           errorMsg(errNum).c_str()));
+      (fmt("Failed to create socket. Cause:%s", errorMsg(errNum).c_str()));
   }
   int sockopt = 1;
   if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
@@ -182,15 +205,15 @@ void SocketCore::create(int family, int protocol)
     errNum = SOCKET_ERRNO;
     CLOSE(fd);
     throw DL_ABORT_EX
-      (fmt("Failed to create socket. Cause:%s",
-           errorMsg(errNum).c_str()));
+      (fmt("Failed to create socket. Cause:%s", errorMsg(errNum).c_str()));
   }
   sockfd_ = fd;
 }
 
-static sock_t bindInternal(int family, int socktype, int protocol,
-                           const struct sockaddr* addr, socklen_t addrlen,
-                           std::string& error)
+static sock_t bindInternal
+(int family, int socktype, int protocol,
+ const struct sockaddr* addr, socklen_t addrlen,
+ std::string& error)
 {
   int errNum;
   sock_t fd = socket(family, socktype, protocol);
@@ -234,7 +257,7 @@ static sock_t bindTo
 {
   struct addrinfo* res;
   int s = callGetaddrinfo(&res, host, util::uitos(port).c_str(),
-                          family, sockType, getaddrinfoFlags, 0);  
+                          family, sockType, getaddrinfoFlags, 0);
   if(s) {
     error = gai_strerror(s);
     return -1;
@@ -264,15 +287,15 @@ void SocketCore::bindWithFamily(uint16_t port, int family, int flags)
 }
 
 void SocketCore::bind
-(const std::string& addr, uint16_t port, int family, int flags)
+(const char* addr, uint16_t port, int family, int flags)
 {
   closeConnection();
   std::string error;
   const char* addrp;
-  if(addr.empty()) {
-    addrp = 0;
+  if(addr && addr[0]) {
+    addrp = addr;
   } else {
-    addrp = addr.c_str();
+    addrp = 0;
   }
   if(!(flags&AI_PASSIVE) || bindAddrs_.empty()) {
     sock_t fd = bindTo(addrp, port, family, sockType_, flags, error);
@@ -280,14 +303,12 @@ void SocketCore::bind
       sockfd_ = fd;
     }
   } else {
-    for(std::vector<std::pair<struct sockaddr_storage, socklen_t> >::
+    for(std::vector<std::pair<sockaddr_union, socklen_t> >::
           const_iterator i = bindAddrs_.begin(), eoi = bindAddrs_.end();
         i != eoi; ++i) {
       char host[NI_MAXHOST];
       int s;
-      s = getnameinfo(reinterpret_cast<const struct sockaddr*>(&(*i).first),
-                      (*i).second,
-                      host, NI_MAXHOST, 0, 0,
+      s = getnameinfo(&(*i).first.sa, (*i).second, host, NI_MAXHOST, 0, 0,
                       NI_NUMERICHOST);
       if(s) {
         error = gai_strerror(s);
@@ -297,7 +318,7 @@ void SocketCore::bind
         error = "Given address and resolved address do not match.";
         continue;
       }
-      sock_t fd = bindTo(addrp, port, family, sockType_, flags, error);
+      sock_t fd = bindTo(host, port, family, sockType_, flags, error);
       if(fd != (sock_t)-1) {
         sockfd_ = fd;
         break;
@@ -311,7 +332,7 @@ void SocketCore::bind
 
 void SocketCore::bind(uint16_t port, int flags)
 {
-  bind(A2STR::NIL, port, protocolFamily_, flags);
+  bind(0, port, protocolFamily_, flags);
 }
 
 void SocketCore::bind(const struct sockaddr* addr, socklen_t addrlen)
@@ -332,37 +353,37 @@ void SocketCore::beginListen()
     int errNum = SOCKET_ERRNO;
     throw DL_ABORT_EX(fmt(EX_SOCKET_LISTEN, errorMsg(errNum).c_str()));
   }
+  setNonBlockingMode();
 }
 
-SocketCore* SocketCore::acceptConnection() const
+SharedHandle<SocketCore> SocketCore::acceptConnection() const
 {
-  struct sockaddr_storage sockaddr;
+  sockaddr_union sockaddr;
   socklen_t len = sizeof(sockaddr);
   sock_t fd;
-  while((fd = accept(sockfd_, reinterpret_cast<struct sockaddr*>(&sockaddr),
-                     &len)) == (sock_t) -1 &&
+  while((fd = accept(sockfd_, &sockaddr.sa, &len)) == (sock_t) -1 &&
         SOCKET_ERRNO == A2_EINTR);
   int errNum = SOCKET_ERRNO;
   if(fd == (sock_t) -1) {
     throw DL_ABORT_EX(fmt(EX_SOCKET_ACCEPT, errorMsg(errNum).c_str()));
   }
-  return new SocketCore(fd, sockType_);
+  SharedHandle<SocketCore> sock(new SocketCore(fd, sockType_));
+  sock->setNonBlockingMode();
+  return sock;
 }
 
-void SocketCore::getAddrInfo(std::pair<std::string, uint16_t>& addrinfo) const
+int SocketCore::getAddrInfo(std::pair<std::string, uint16_t>& addrinfo) const
 {
-  struct sockaddr_storage sockaddr;
+  sockaddr_union sockaddr;
   socklen_t len = sizeof(sockaddr);
   getAddrInfo(sockaddr, len);
-  addrinfo = util::getNumericNameInfo
-    (reinterpret_cast<const struct sockaddr*>(&sockaddr), len);
+  addrinfo = util::getNumericNameInfo(&sockaddr.sa, len);
+  return sockaddr.storage.ss_family;
 }
 
-void SocketCore::getAddrInfo
-(struct sockaddr_storage& sockaddr, socklen_t& len) const
+void SocketCore::getAddrInfo(sockaddr_union& sockaddr, socklen_t& len) const
 {
-  struct sockaddr* addrp = reinterpret_cast<struct sockaddr*>(&sockaddr);
-  if(getsockname(sockfd_, addrp, &len) == -1) {
+  if(getsockname(sockfd_, &sockaddr.sa, &len) == -1) {
     int errNum = SOCKET_ERRNO;
     throw DL_ABORT_EX(fmt(EX_SOCKET_GET_NAME, errorMsg(errNum).c_str()));
   }
@@ -370,22 +391,22 @@ void SocketCore::getAddrInfo
 
 int SocketCore::getAddressFamily() const
 {
-  struct sockaddr_storage sockaddr;
+  sockaddr_union sockaddr;
   socklen_t len = sizeof(sockaddr);
   getAddrInfo(sockaddr, len);
-  return sockaddr.ss_family;
+  return sockaddr.storage.ss_family;
 }
 
-void SocketCore::getPeerInfo(std::pair<std::string, uint16_t>& peerinfo) const
+int SocketCore::getPeerInfo(std::pair<std::string, uint16_t>& peerinfo) const
 {
-  struct sockaddr_storage sockaddr;
+  sockaddr_union sockaddr;
   socklen_t len = sizeof(sockaddr);
-  struct sockaddr* addrp = reinterpret_cast<struct sockaddr*>(&sockaddr);
-  if(getpeername(sockfd_, addrp, &len) == -1) {
+  if(getpeername(sockfd_, &sockaddr.sa, &len) == -1) {
     int errNum = SOCKET_ERRNO;
     throw DL_ABORT_EX(fmt(EX_SOCKET_GET_NAME, errorMsg(errNum).c_str()));
   }
-  peerinfo = util::getNumericNameInfo(addrp, len);
+  peerinfo = util::getNumericNameInfo(&sockaddr.sa, len);
+  return sockaddr.storage.ss_family;
 }
 
 void SocketCore::establishConnection(const std::string& host, uint16_t port)
@@ -397,9 +418,7 @@ void SocketCore::establishConnection(const std::string& host, uint16_t port)
   s = callGetaddrinfo(&res, host.c_str(), util::uitos(port).c_str(),
                       protocolFamily_, sockType_, 0, 0);
   if(s) {
-    throw DL_ABORT_EX(fmt(EX_RESOLVE_HOSTNAME,
-                          host.c_str(),
-                          gai_strerror(s)));
+    throw DL_ABORT_EX(fmt(EX_RESOLVE_HOSTNAME, host.c_str(), gai_strerror(s)));
   }
   WSAAPI_AUTO_DELETE<struct addrinfo*> resDeleter(res, freeaddrinfo);
   struct addrinfo* rp;
@@ -421,11 +440,10 @@ void SocketCore::establishConnection(const std::string& host, uint16_t port)
     }
     if(!bindAddrs_.empty()) {
       bool bindSuccess = false;
-      for(std::vector<std::pair<struct sockaddr_storage, socklen_t> >::
+      for(std::vector<std::pair<sockaddr_union, socklen_t> >::
             const_iterator i = bindAddrs_.begin(), eoi = bindAddrs_.end();
           i != eoi; ++i) {
-        if(::bind(fd,reinterpret_cast<const struct sockaddr*>(&(*i).first),
-                  (*i).second) == -1) {
+        if(::bind(fd, &(*i).first.sa, (*i).second) == -1) {
           errNum = SOCKET_ERRNO;
           error = errorMsg(errNum);
           A2_LOG_DEBUG(fmt(EX_SOCKET_BIND, error.c_str()));
@@ -456,10 +474,7 @@ void SocketCore::establishConnection(const std::string& host, uint16_t port)
     break;
   }
   if(sockfd_ == (sock_t) -1) {
-    throw DL_ABORT_EX
-      (fmt(EX_SOCKET_CONNECT,
-           host.c_str(),
-           error.c_str()));
+    throw DL_ABORT_EX(fmt(EX_SOCKET_CONNECT, host.c_str(), error.c_str()));
   }
 }
 
@@ -470,7 +485,7 @@ void SocketCore::setSockOpt
     int errNum = SOCKET_ERRNO;
     throw DL_ABORT_EX(fmt(EX_SOCKET_SET_OPT, errorMsg(errNum).c_str()));
   }
-}   
+}
 
 void SocketCore::setMulticastInterface(const std::string& localAddr)
 {
@@ -478,8 +493,9 @@ void SocketCore::setMulticastInterface(const std::string& localAddr)
   if(localAddr.empty()) {
     addr.s_addr = htonl(INADDR_ANY);
   } else {
-    if(inet_aton(localAddr.c_str(), &addr) == 0) {
-      throw DL_ABORT_EX(fmt("inet_aton failed for %s", localAddr.c_str()));
+    if(inetPton(AF_INET, localAddr.c_str(), &addr) != 0) {
+      throw DL_ABORT_EX(fmt("%s is not valid IPv4 numeric address",
+                            localAddr.c_str()));
     }
   }
   setSockOpt(IPPROTO_IP, IP_MULTICAST_IF, &addr, sizeof(addr));
@@ -500,15 +516,17 @@ void SocketCore::joinMulticastGroup
  const std::string& localAddr)
 {
   in_addr multiAddr;
-  if(inet_aton(multicastAddr.c_str(), &multiAddr) == 0) {
-    throw DL_ABORT_EX(fmt("inet_aton failed for %s", multicastAddr.c_str()));
+  if(inetPton(AF_INET, multicastAddr.c_str(), &multiAddr) != 0) {
+    throw DL_ABORT_EX(fmt("%s is not valid IPv4 numeric address",
+                          multicastAddr.c_str()));
   }
   in_addr ifAddr;
   if(localAddr.empty()) {
     ifAddr.s_addr = htonl(INADDR_ANY);
   } else {
-    if(inet_aton(localAddr.c_str(), &ifAddr) == 0) {
-      throw DL_ABORT_EX(fmt("inet_aton failed for %s", localAddr.c_str()));
+    if(inetPton(AF_INET, localAddr.c_str(), &ifAddr) != 0) {
+      throw DL_ABORT_EX(fmt("%s is not valid IPv4 numeric address",
+                            localAddr.c_str()));
     }
   }
   struct ip_mreq mreq;
@@ -518,6 +536,12 @@ void SocketCore::joinMulticastGroup
   setSockOpt(IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq));
 }
 
+void SocketCore::setTcpNodelay(bool f)
+{
+  int val = f;
+  setSockOpt(IPPROTO_TCP, TCP_NODELAY, &val, sizeof(val));
+}
+
 void SocketCore::setNonBlockingMode()
 {
 #ifdef __MINGW32__
@@ -632,8 +656,7 @@ bool SocketCore::isWritable(time_t timeout)
       return false;
     } else {
       throw DL_RETRY_EX
-        (fmt(EX_SOCKET_CHECK_WRITABLE,
-             errorMsg(errNum).c_str()));
+        (fmt(EX_SOCKET_CHECK_WRITABLE, errorMsg(errNum).c_str()));
     }
   }
 #endif // !HAVE_POLL
@@ -679,8 +702,7 @@ bool SocketCore::isReadable(time_t timeout)
       return false;
     } else {
       throw DL_RETRY_EX
-        (fmt(EX_SOCKET_CHECK_READABLE,
-             errorMsg(errNum).c_str()));
+        (fmt(EX_SOCKET_CHECK_READABLE, errorMsg(errNum).c_str()));
     }
   }
 #endif // !HAVE_POLL
@@ -714,14 +736,57 @@ void SocketCore::gnutlsRecordCheckDirection()
 }
 #endif // HAVE_LIBGNUTLS
 
-ssize_t SocketCore::writeData(const char* data, size_t len)
+ssize_t SocketCore::writeVector(a2iovec *iov, size_t iovcnt)
+{
+  ssize_t ret = 0;
+  wantRead_ = false;
+  wantWrite_ = false;
+  if(!secure_) {
+#ifdef __MINGW32__
+    DWORD nsent;
+    int rv = WSASend(sockfd_, iov, iovcnt, &nsent, 0, 0, 0);
+    if(rv == 0) {
+      ret = nsent;
+    } else {
+      ret = -1;
+    }
+#else // !__MINGW32__
+    while((ret = writev(sockfd_, iov, iovcnt)) == -1 &&
+          SOCKET_ERRNO == A2_EINTR);
+#endif // !__MINGW32__
+    int errNum = SOCKET_ERRNO;
+    if(ret == -1) {
+      if(A2_WOULDBLOCK(errNum)) {
+        wantWrite_ = true;
+        ret = 0;
+      } else {
+        throw DL_RETRY_EX(fmt(EX_SOCKET_SEND, errorMsg(errNum).c_str()));
+      }
+    }
+  } else {
+    // For SSL/TLS, we could not use writev, so just iterate vector
+    // and write the data in normal way.
+    for(size_t i = 0; i < iovcnt; ++i) {
+      ssize_t rv = writeData(iov[i].A2IOVEC_BASE, iov[i].A2IOVEC_LEN);
+      if(rv == 0) {
+        break;
+      }
+      ret += rv;
+    }
+  }
+  return ret;
+}
+
+ssize_t SocketCore::writeData(const void* data, size_t len)
 {
   ssize_t ret = 0;
   wantRead_ = false;
   wantWrite_ = false;
 
   if(!secure_) {
-    while((ret = send(sockfd_, data, len, 0)) == -1 && SOCKET_ERRNO == A2_EINTR);
+    // Cast for Windows send()
+    while((ret = send(sockfd_, reinterpret_cast<const char*>(data),
+                      len, 0)) == -1 && SOCKET_ERRNO == A2_EINTR);
     int errNum = SOCKET_ERRNO;
     if(ret == -1) {
       if(A2_WOULDBLOCK(errNum)) {
@@ -733,14 +798,14 @@ ssize_t SocketCore::writeData(const char* data, size_t len)
     }
   } else {
 #ifdef HAVE_OPENSSL
+    ERR_clear_error();
     ret = SSL_write(ssl, data, len);
     if(ret < 0) {
       ret = sslHandleEAGAIN(ret);
     }
     if(ret < 0) {
       throw DL_RETRY_EX
-        (fmt(EX_SOCKET_SEND,
-             ERR_error_string(SSL_get_error(ssl, ret), 0)));
+        (fmt(EX_SOCKET_SEND, ERR_error_string(ERR_get_error(), 0)));
     }
 #endif // HAVE_OPENSSL
 #ifdef HAVE_LIBGNUTLS
@@ -758,15 +823,16 @@ ssize_t SocketCore::writeData(const char* data, size_t len)
   return ret;
 }
 
-void SocketCore::readData(char* data, size_t& len)
+void SocketCore::readData(void* data, size_t& len)
 {
   ssize_t ret = 0;
   wantRead_ = false;
   wantWrite_ = false;
 
-  if(!secure_) {    
-    while((ret = recv(sockfd_, data, len, 0)) == -1 && SOCKET_ERRNO == A2_EINTR);
-    
+  if(!secure_) {
+    // Cast for Windows recv()
+    while((ret = recv(sockfd_, reinterpret_cast<char*>(data), len, 0)) == -1 &&
+          SOCKET_ERRNO == A2_EINTR);
     int errNum = SOCKET_ERRNO;
     if(ret == -1) {
       if(A2_WOULDBLOCK(errNum)) {
@@ -780,14 +846,14 @@ void SocketCore::readData(char* data, size_t& len)
 #ifdef HAVE_OPENSSL
     // for SSL
     // TODO handling len == 0 case required
+    ERR_clear_error();
     ret = SSL_read(ssl, data, len);
     if(ret < 0) {
       ret = sslHandleEAGAIN(ret);
     }
     if(ret < 0) {
       throw DL_RETRY_EX
-        (fmt(EX_SOCKET_RECV,
-             ERR_error_string(SSL_get_error(ssl, ret), 0)));
+        (fmt(EX_SOCKET_RECV, ERR_error_string(ERR_get_error(), 0)));
     }
 #endif // HAVE_OPENSSL
 #ifdef HAVE_LIBGNUTLS
@@ -805,58 +871,59 @@ void SocketCore::readData(char* data, size_t& len)
   len = ret;
 }
 
-void SocketCore::prepareSecureConnection()
+#ifdef ENABLE_SSL
+
+bool SocketCore::tlsAccept()
 {
-  if(!secure_) {
+  return tlsHandshake(svTlsContext_.get(), A2STR::NIL);
+}
+
+bool SocketCore::tlsConnect(const std::string& hostname)
+{
+  return tlsHandshake(clTlsContext_.get(), hostname);
+}
+
+bool SocketCore::tlsHandshake(TLSContext* tlsctx, const std::string& hostname)
+{
+  wantRead_ = false;
+  wantWrite_ = false;
 #ifdef HAVE_OPENSSL
-    // for SSL
-    ssl = SSL_new(tlsContext_->getSSLCtx());
+  switch(secure_) {
+  case A2_TLS_NONE:
+    ssl = SSL_new(tlsctx->getSSLCtx());
     if(!ssl) {
       throw DL_ABORT_EX
-        (fmt(EX_SSL_INIT_FAILURE,
-             ERR_error_string(ERR_get_error(), 0)));
+        (fmt(EX_SSL_INIT_FAILURE, ERR_error_string(ERR_get_error(), 0)));
     }
     if(SSL_set_fd(ssl, sockfd_) == 0) {
       throw DL_ABORT_EX
-        (fmt(EX_SSL_INIT_FAILURE,
-             ERR_error_string(ERR_get_error(), 0)));
+        (fmt(EX_SSL_INIT_FAILURE, ERR_error_string(ERR_get_error(), 0)));
     }
-#endif // HAVE_OPENSSL
-#ifdef HAVE_LIBGNUTLS
-    int r;
-    gnutls_init(&sslSession_, GNUTLS_CLIENT);
-    // It seems err is not error message, but the argument string
-    // which causes syntax error.
-    const char* err;
-    // Disables TLS1.1 here because there are servers that don't
-    // understand TLS1.1.
-    r = gnutls_priority_set_direct(sslSession_, "NORMAL:!VERS-TLS1.1", &err);
-    if(r != GNUTLS_E_SUCCESS) {
-      throw DL_ABORT_EX(fmt(EX_SSL_INIT_FAILURE, gnutls_strerror(r)));
+    // Fall through
+#ifdef SSL_CTRL_SET_TLSEXT_HOSTNAME
+    if(tlsctx->getSide() == TLS_CLIENT && !util::isNumericHost(hostname)) {
+      // TLS extensions: SNI.  There is not documentation about the
+      // return code for this function (actually this is macro
+      // wrapping SSL_ctrl at the time of this writing).
+      SSL_set_tlsext_host_name(ssl, hostname.c_str());
+    }
+#endif // SSL_CTRL_SET_TLSEXT_HOSTNAME
+    secure_ = A2_TLS_HANDSHAKING;
+    // Fall through
+  case A2_TLS_HANDSHAKING: {
+    ERR_clear_error();
+    int e;
+    if(tlsctx->getSide() == TLS_CLIENT) {
+      e = SSL_connect(ssl);
+    } else {
+      e = SSL_accept(ssl);
     }
-    // put the x509 credentials to the current session
-    gnutls_credentials_set(sslSession_, GNUTLS_CRD_CERTIFICATE,
-                           tlsContext_->getCertCred());
-    gnutls_transport_set_ptr(sslSession_, (gnutls_transport_ptr_t)sockfd_);
-#endif // HAVE_LIBGNUTLS
-    secure_ = 1;
-  }
-}
-
-bool SocketCore::initiateSecureConnection(const std::string& hostname)
-{
-  if(secure_ == 1) {
-    wantRead_ = false;
-    wantWrite_ = false;
-#ifdef HAVE_OPENSSL
-    int e = SSL_connect(ssl);
 
     if (e <= 0) {
       int ssl_error = SSL_get_error(ssl, e);
       switch(ssl_error) {
       case SSL_ERROR_NONE:
         break;
-
       case SSL_ERROR_WANT_READ:
         wantRead_ = true;
         return false;
@@ -870,9 +937,21 @@ bool SocketCore::initiateSecureConnection(const std::string& hostname)
         }
         break;
 
-      case SSL_ERROR_SYSCALL:
-        throw DL_ABORT_EX(EX_SSL_IO_ERROR);
-
+      case SSL_ERROR_SYSCALL: {
+        int sslErr = ERR_get_error();
+        if(sslErr == 0) {
+          if(e == 0) {
+            throw DL_ABORT_EX("Got EOF in SSL handshake");
+          } else if(e == -1) {
+            throw DL_ABORT_EX(fmt("SSL I/O error: %s", strerror(errno)));
+          } else {
+            throw DL_ABORT_EX(EX_SSL_IO_ERROR);
+          }
+        } else {
+          throw DL_ABORT_EX(fmt("SSL I/O error: %s",
+                                ERR_error_string(sslErr, 0)));
+        }
+      }
       case SSL_ERROR_SSL:
         throw DL_ABORT_EX(EX_SSL_PROTOCOL_ERROR);
 
@@ -880,7 +959,8 @@ bool SocketCore::initiateSecureConnection(const std::string& hostname)
         throw DL_ABORT_EX(fmt(EX_SSL_UNKNOWN_ERROR, ssl_error));
       }
     }
-    if(tlsContext_->peerVerificationEnabled()) {
+    if(tlsctx->getSide() == TLS_CLIENT &&
+       tlsctx->peerVerificationEnabled()) {
       // verify peer
       X509* peerCert = SSL_get_peer_certificate(ssl);
       if(!peerCert) {
@@ -894,37 +974,113 @@ bool SocketCore::initiateSecureConnection(const std::string& hostname)
           (fmt(MSG_CERT_VERIFICATION_FAILED,
                X509_verify_cert_error_string(verifyResult)));
       }
-      X509_NAME* name = X509_get_subject_name(peerCert);
-      if(!name) {
-        throw DL_ABORT_EX("Could not get X509 name object from the certificate.");
+      std::string commonName;
+      std::vector<std::string> dnsNames;
+      std::vector<std::string> ipAddrs;
+      GENERAL_NAMES* altNames;
+      altNames = reinterpret_cast<GENERAL_NAMES*>
+        (X509_get_ext_d2i(peerCert, NID_subject_alt_name, NULL, NULL));
+      if(altNames) {
+        auto_delete<GENERAL_NAMES*> altNamesDeleter
+          (altNames, GENERAL_NAMES_free);
+        size_t n = sk_GENERAL_NAME_num(altNames);
+        for(size_t i = 0; i < n; ++i) {
+          const GENERAL_NAME* altName = sk_GENERAL_NAME_value(altNames, i);
+          if(altName->type == GEN_DNS) {
+            const char* name =
+              reinterpret_cast<char*>(ASN1_STRING_data(altName->d.ia5));
+            if(!name) {
+              continue;
+            }
+            size_t len = ASN1_STRING_length(altName->d.ia5);
+            dnsNames.push_back(std::string(name, len));
+          } else if(altName->type == GEN_IPADD) {
+            const unsigned char* ipAddr = altName->d.iPAddress->data;
+            if(!ipAddr) {
+              continue;
+            }
+            size_t len = altName->d.iPAddress->length;
+            ipAddrs.push_back(std::string(reinterpret_cast<const char*>(ipAddr),
+                                          len));
+          }
+        }
+      }
+      X509_NAME* subjectName = X509_get_subject_name(peerCert);
+      if(!subjectName) {
+        throw DL_ABORT_EX
+          ("Could not get X509 name object from the certificate.");
       }
-
-      bool hostnameOK = false;
       int lastpos = -1;
-      while(true) {
-        lastpos = X509_NAME_get_index_by_NID(name, NID_commonName, lastpos);
+      while(1) {
+        lastpos = X509_NAME_get_index_by_NID(subjectName, NID_commonName,
+                                             lastpos);
         if(lastpos == -1) {
           break;
         }
-        X509_NAME_ENTRY* entry = X509_NAME_get_entry(name, lastpos);
+        X509_NAME_ENTRY* entry = X509_NAME_get_entry(subjectName, lastpos);
         unsigned char* out;
-        int outlen = ASN1_STRING_to_UTF8(&out, X509_NAME_ENTRY_get_data(entry));
+        int outlen = ASN1_STRING_to_UTF8(&out,
+                                         X509_NAME_ENTRY_get_data(entry));
         if(outlen < 0) {
           continue;
         }
-        std::string commonName(&out[0], &out[outlen]);
+        commonName.assign(&out[0], &out[outlen]);
         OPENSSL_free(out);
-        if(commonName == hostname) {
-          hostnameOK = true;
-          break;
-        }
+        break;
       }
-      if(!hostnameOK) {
+      if(!net::verifyHostname(hostname, dnsNames, ipAddrs, commonName)) {
         throw DL_ABORT_EX(MSG_HOSTNAME_NOT_MATCH);
       }
     }
+    secure_ = A2_TLS_CONNECTED;
+    break;
+  }
+  default:
+    break;
+  }
 #endif // HAVE_OPENSSL
 #ifdef HAVE_LIBGNUTLS
+  switch(secure_) {
+  case A2_TLS_NONE:
+    int r;
+    gnutls_init(&sslSession_,
+                tlsctx->getSide() == TLS_CLIENT ?
+                GNUTLS_CLIENT : GNUTLS_SERVER);
+    // It seems err is not error message, but the argument string
+    // which causes syntax error.
+    const char* err;
+    // For client side, disables TLS1.1 here because there are servers
+    // that don't understand TLS1.1.  TODO Is this still necessary?
+    r = gnutls_priority_set_direct(sslSession_,
+                                   tlsctx->getSide() == TLS_CLIENT ?
+                                   "NORMAL:-VERS-TLS1.1" :
+                                   "NORMAL",
+                                   &err);
+    if(r != GNUTLS_E_SUCCESS) {
+      throw DL_ABORT_EX(fmt(EX_SSL_INIT_FAILURE, gnutls_strerror(r)));
+    }
+    // put the x509 credentials to the current session
+    gnutls_credentials_set(sslSession_, GNUTLS_CRD_CERTIFICATE,
+                           tlsctx->getCertCred());
+    gnutls_transport_set_ptr(sslSession_, (gnutls_transport_ptr_t)sockfd_);
+    if(tlsctx->getSide() == TLS_CLIENT) {
+      // Check hostname is not numeric and it includes ".". Setting
+      // "localhost" will produce TLS alert.
+      if(!util::isNumericHost(hostname) &&
+         hostname.find(".") != std::string::npos) {
+        // TLS extensions: SNI
+        int ret = gnutls_server_name_set(sslSession_, GNUTLS_NAME_DNS,
+                                         hostname.c_str(), hostname.size());
+        if(ret < 0) {
+          A2_LOG_WARN(fmt
+                      ("Setting hostname in SNI extension failed. Cause: %s",
+                       gnutls_strerror(ret)));
+        }
+      }
+    }
+    secure_ = A2_TLS_HANDSHAKING;
+    // Fall through
+  case A2_TLS_HANDSHAKING: {
     int ret = gnutls_handshake(sslSession_);
     if(ret == GNUTLS_E_AGAIN) {
       gnutlsRecordCheckDirection();
@@ -933,7 +1089,7 @@ bool SocketCore::initiateSecureConnection(const std::string& hostname)
       throw DL_ABORT_EX(fmt(EX_SSL_INIT_FAILURE, gnutls_strerror(ret)));
     }
 
-    if(tlsContext_->peerVerificationEnabled()) {
+    if(tlsctx->getSide() == TLS_CLIENT && tlsctx->peerVerificationEnabled()) {
       // verify peer
       unsigned int status;
       ret = gnutls_certificate_verify_peers2(sslSession_, &status);
@@ -953,6 +1109,17 @@ bool SocketCore::initiateSecureConnection(const std::string& hostname)
         if(status & GNUTLS_CERT_SIGNER_NOT_FOUND) {
           errors += " `issuer is not known'";
         }
+        // TODO should check GNUTLS_CERT_SIGNER_NOT_CA ?
+        if(status & GNUTLS_CERT_INSECURE_ALGORITHM) {
+          errors += " `insecure algorithm'";
+        }
+        if(status & GNUTLS_CERT_NOT_ACTIVATED) {
+          errors += " `not activated yet'";
+        }
+        if(status & GNUTLS_CERT_EXPIRED) {
+          errors += " `expired'";
+        }
+        // TODO Add GNUTLS_CERT_SIGNATURE_FAILURE here
         if(!errors.empty()) {
           throw DL_ABORT_EX(fmt(MSG_CERT_VERIFICATION_FAILED, errors.c_str()));
         }
@@ -965,7 +1132,7 @@ bool SocketCore::initiateSecureConnection(const std::string& hostname)
       unsigned int peerCertsLength;
       const gnutls_datum_t* peerCerts = gnutls_certificate_get_peers
         (sslSession_, &peerCertsLength);
-      if(!peerCerts) {
+      if(!peerCerts || peerCertsLength == 0 ) {
         throw DL_ABORT_EX(MSG_NO_CERT_FOUND);
       }
       Time now;
@@ -986,7 +1153,30 @@ bool SocketCore::initiateSecureConnection(const std::string& hostname)
                  gnutls_strerror(ret)));
         }
         if(i == 0) {
-          if(!gnutls_x509_crt_check_hostname(cert, hostname.c_str())) {
+          std::string commonName;
+          std::vector<std::string> dnsNames;
+          std::vector<std::string> ipAddrs;
+          int ret = 0;
+          char altName[256];
+          size_t altNameLen;
+          for(int j = 0; !(ret < 0); ++j) {
+            altNameLen = sizeof(altName);
+            ret = gnutls_x509_crt_get_subject_alt_name(cert, j, altName,
+                                                       &altNameLen, 0);
+            if(ret == GNUTLS_SAN_DNSNAME) {
+              dnsNames.push_back(std::string(altName, altNameLen));
+            } else if(ret == GNUTLS_SAN_IPADDRESS) {
+              ipAddrs.push_back(std::string(altName, altNameLen));
+            }
+          }
+          altNameLen = sizeof(altName);
+          ret = gnutls_x509_crt_get_dn_by_oid(cert,
+                                              GNUTLS_OID_X520_COMMON_NAME, 0, 0,
+                                              altName, &altNameLen);
+          if(ret == 0) {
+            commonName.assign(altName, altNameLen);
+          }
+          if(!net::verifyHostname(hostname, dnsNames, ipAddrs, commonName)) {
             throw DL_ABORT_EX(MSG_HOSTNAME_NOT_MATCH);
           }
         }
@@ -1006,15 +1196,19 @@ bool SocketCore::initiateSecureConnection(const std::string& hostname)
         }
       }
     }
-#endif // HAVE_LIBGNUTLS
-    secure_ = 2;
-    return true;
-  } else {
-    return true;
+    secure_ = A2_TLS_CONNECTED;
+    break;
   }
+  default:
+    break;
+  }
+#endif // HAVE_LIBGNUTLS
+  return true;
 }
 
-ssize_t SocketCore::writeData(const char* data, size_t len,
+#endif // ENABLE_SSL
+
+ssize_t SocketCore::writeData(const void* data, size_t len,
                               const std::string& host, uint16_t port)
 {
   wantRead_ = false;
@@ -1032,7 +1226,9 @@ ssize_t SocketCore::writeData(const char* data, size_t len,
   ssize_t r = -1;
   int errNum = 0;
   for(rp = res; rp; rp = rp->ai_next) {
-    while((r = sendto(sockfd_, data, len, 0, rp->ai_addr, rp->ai_addrlen)) == -1
+    // Cast for Windows sendto()
+    while((r = sendto(sockfd_, reinterpret_cast<const char*>(data), len, 0,
+                      rp->ai_addr, rp->ai_addrlen)) == -1
           && A2_EINTR == SOCKET_ERRNO);
     errNum = SOCKET_ERRNO;
     if(r == static_cast<ssize_t>(len)) {
@@ -1050,18 +1246,19 @@ ssize_t SocketCore::writeData(const char* data, size_t len,
   return r;
 }
 
-ssize_t SocketCore::readDataFrom(char* data, size_t len,
+ssize_t SocketCore::readDataFrom(void* data, size_t len,
                                  std::pair<std::string /* numerichost */,
                                  uint16_t /* port */>& sender)
 {
   wantRead_ = false;
   wantWrite_ = false;
-  struct sockaddr_storage sockaddr;
-  socklen_t sockaddrlen = sizeof(struct sockaddr_storage);
-  struct sockaddr* addrp = reinterpret_cast<struct sockaddr*>(&sockaddr);
+  sockaddr_union sockaddr;
+  socklen_t sockaddrlen = sizeof(sockaddr);
   ssize_t r;
-  while((r = recvfrom(sockfd_, data, len, 0, addrp, &sockaddrlen)) == -1 &&
-        A2_EINTR == SOCKET_ERRNO);
+  // Cast for Windows recvfrom()
+  while((r = recvfrom(sockfd_, reinterpret_cast<char*>(data), len, 0,
+                      &sockaddr.sa, &sockaddrlen)) == -1
+        && A2_EINTR == SOCKET_ERRNO);
   int errNum = SOCKET_ERRNO;
   if(r == -1) {
     if(A2_WOULDBLOCK(errNum)) {
@@ -1071,7 +1268,7 @@ ssize_t SocketCore::readDataFrom(char* data, size_t len,
       throw DL_RETRY_EX(fmt(EX_SOCKET_RECV, errorMsg(errNum).c_str()));
     }
   } else {
-    sender = util::getNumericNameInfo(addrp, sockaddrlen);
+    sender = util::getNumericNameInfo(&sockaddr.sa, sockaddrlen);
   }
 
   return r;
@@ -1086,8 +1283,7 @@ std::string SocketCore::getSocketError() const
                 (a2_sockopt_t) &error, &optlen) == -1) {
     int errNum = SOCKET_ERRNO;
     throw DL_ABORT_EX
-      (fmt("Failed to get socket error: %s",
-           errorMsg(errNum).c_str()));
+      (fmt("Failed to get socket error: %s", errorMsg(errNum).c_str()));
   }
   if(error != 0) {
     return errorMsg(error);
@@ -1108,23 +1304,19 @@ bool SocketCore::wantWrite() const
 
 void SocketCore::bindAddress(const std::string& iface)
 {
-  std::vector<std::pair<struct sockaddr_storage, socklen_t> > bindAddrs;
+  std::vector<std::pair<sockaddr_union, socklen_t> > bindAddrs;
   getInterfaceAddress(bindAddrs, iface, protocolFamily_);
   if(bindAddrs.empty()) {
     throw DL_ABORT_EX
-      (fmt(MSG_INTERFACE_NOT_FOUND,
-           iface.c_str(),
-           "not available"));
+      (fmt(MSG_INTERFACE_NOT_FOUND, iface.c_str(), "not available"));
   } else {
-    bindAddrs_ = bindAddrs;
-    for(std::vector<std::pair<struct sockaddr_storage, socklen_t> >::
+    bindAddrs_.swap(bindAddrs);
+    for(std::vector<std::pair<sockaddr_union, socklen_t> >::
           const_iterator i = bindAddrs_.begin(), eoi = bindAddrs_.end();
         i != eoi; ++i) {
       char host[NI_MAXHOST];
       int s;
-      s = getnameinfo(reinterpret_cast<const struct sockaddr*>(&(*i).first),
-                      (*i).second,
-                      host, NI_MAXHOST, 0, 0,
+      s = getnameinfo(&(*i).first.sa, (*i).second, host, NI_MAXHOST, 0, 0,
                       NI_NUMERICHOST);
       if(s == 0) {
         A2_LOG_DEBUG(fmt("Sockets will bind to %s", host));
@@ -1134,7 +1326,7 @@ void SocketCore::bindAddress(const std::string& iface)
 }
 
 void getInterfaceAddress
-(std::vector<std::pair<struct sockaddr_storage, socklen_t> >& ifAddrs,
+(std::vector<std::pair<sockaddr_union, socklen_t> >& ifAddrs,
  const std::string& iface, int family, int aiFlags)
 {
   A2_LOG_DEBUG(fmt("Finding interface %s", iface.c_str()));
@@ -1146,8 +1338,8 @@ void getInterfaceAddress
     A2_LOG_INFO(fmt(MSG_INTERFACE_NOT_FOUND,
                     iface.c_str(), errorMsg(errNum).c_str()));
   } else {
-    auto_delete<struct ifaddrs*> ifaddrDeleter(ifaddr, freeifaddrs);
-    for(struct ifaddrs* ifa = ifaddr; ifa; ifa = ifa->ifa_next) {
+    auto_delete<ifaddrs*> ifaddrDeleter(ifaddr, freeifaddrs);
+    for(ifaddrs* ifa = ifaddr; ifa; ifa = ifa->ifa_next) {
       if(!ifa->ifa_addr) {
         continue;
       }
@@ -1167,38 +1359,36 @@ void getInterfaceAddress
       } else {
         continue;
       }
-      if(std::string(ifa->ifa_name) == iface) {
-        socklen_t bindAddrLen = iffamily == AF_INET?sizeof(struct sockaddr_in):
-          sizeof(struct sockaddr_in6);
-        struct sockaddr_storage bindAddr;
+      if(strcmp(iface.c_str(), ifa->ifa_name) == 0) {
+        socklen_t bindAddrLen =
+          iffamily == AF_INET ? sizeof(sockaddr_in) : sizeof(sockaddr_in6);
+        sockaddr_union bindAddr;
         memset(&bindAddr, 0, sizeof(bindAddr));
-        memcpy(&bindAddr, ifa->ifa_addr, bindAddrLen);
+        memcpy(&bindAddr.storage, ifa->ifa_addr, bindAddrLen);
         ifAddrs.push_back(std::make_pair(bindAddr, bindAddrLen));
       }
     }
   }
 #endif // HAVE_GETIFADDRS
   if(ifAddrs.empty()) {
-    struct addrinfo* res;
+    addrinfo* res;
     int s;
     s = callGetaddrinfo(&res, iface.c_str(), 0, family, SOCK_STREAM, aiFlags,0);
     if(s) {
       A2_LOG_INFO(fmt(MSG_INTERFACE_NOT_FOUND, iface.c_str(), gai_strerror(s)));
     } else {
-      WSAAPI_AUTO_DELETE<struct addrinfo*> resDeleter(res, freeaddrinfo);
-      struct addrinfo* rp;
+      WSAAPI_AUTO_DELETE<addrinfo*> resDeleter(res, freeaddrinfo);
+      addrinfo* rp;
       for(rp = res; rp; rp = rp->ai_next) {
-        socklen_t bindAddrLen = rp->ai_addrlen;
-        struct sockaddr_storage bindAddr;
-        memset(&bindAddr, 0, sizeof(bindAddr));
-        memcpy(&bindAddr, rp->ai_addr, rp->ai_addrlen);
         // Try to bind socket with this address. If it fails, the
         // address is not for this machine.
         try {
           SocketCore socket;
-          socket.bind
-            (reinterpret_cast<const struct sockaddr*>(&bindAddr), bindAddrLen);
-          ifAddrs.push_back(std::make_pair(bindAddr, bindAddrLen));
+          socket.bind(rp->ai_addr, rp->ai_addrlen);
+          sockaddr_union bindAddr;
+          memset(&bindAddr, 0, sizeof(bindAddr));
+          memcpy(&bindAddr.storage, rp->ai_addr, rp->ai_addrlen);
+          ifAddrs.push_back(std::make_pair(bindAddr, rp->ai_addrlen));
         } catch(RecoverableException& e) {
           continue;
         }
@@ -1234,25 +1424,29 @@ int callGetaddrinfo
   hints.ai_flags = getDefaultAIFlags();
   hints.ai_flags |= flags;
   hints.ai_protocol = protocol;
-  return getaddrinfo(host, service, &hints, resPtr);  
+  return getaddrinfo(host, service, &hints, resPtr);
 }
 
 int inetNtop(int af, const void* src, char* dst, socklen_t size)
 {
   int s;
+  sockaddr_union su;
+  memset(&su, 0, sizeof(su));
   if(af == AF_INET) {
-    sockaddr_in sa;
-    memset(&sa, 0, sizeof(sa));
-    sa.sin_family = AF_INET;
-    memcpy(&sa.sin_addr, src, sizeof(in_addr));
-    s = getnameinfo(reinterpret_cast<const sockaddr*>(&sa), sizeof(sa),
+    su.in.sin_family = AF_INET;
+#ifdef HAVE_SOCKADDR_IN_SIN_LEN
+    su.in.sin_len = sizeof(su.in);
+#endif // HAVE_SOCKADDR_IN_SIN_LEN
+    memcpy(&su.in.sin_addr, src, sizeof(su.in.sin_addr));
+    s = getnameinfo(&su.sa, sizeof(su.in),
                     dst, size, 0, 0, NI_NUMERICHOST);
   } else if(af == AF_INET6) {
-    sockaddr_in6 sa;
-    memset(&sa, 0, sizeof(sa));
-    sa.sin6_family = AF_INET6;
-    memcpy(&sa.sin6_addr, src, sizeof(in6_addr));
-    s = getnameinfo(reinterpret_cast<const sockaddr*>(&sa), sizeof(sa),
+    su.in6.sin6_family = AF_INET6;
+#ifdef HAVE_SOCKADDR_IN6_SIN6_LEN
+    su.in6.sin6_len = sizeof(su.in6);
+#endif // HAVE_SOCKADDR_IN6_SIN6_LEN
+    memcpy(&su.in6.sin6_addr, src, sizeof(su.in6.sin6_addr));
+    s = getnameinfo(&su.sa, sizeof(su.in6),
                     dst, size, 0, 0, NI_NUMERICHOST);
   } else {
     s = EAI_FAMILY;
@@ -1260,4 +1454,94 @@ int inetNtop(int af, const void* src, char* dst, socklen_t size)
   return s;
 }
 
+int inetPton(int af, const char* src, void* dst)
+{
+  union {
+    uint32_t ipv4_addr;
+    unsigned char ipv6_addr[16];
+  } binaddr;
+  size_t len = net::getBinAddr(binaddr.ipv6_addr, src);
+  if(af == AF_INET) {
+    if(len != 4) {
+      return -1;
+    }
+    in_addr* addr = reinterpret_cast<in_addr*>(dst);
+    addr->s_addr = binaddr.ipv4_addr;
+  } else if(af == AF_INET6) {
+    if(len != 16) {
+      return -1;
+    }
+    in6_addr* addr = reinterpret_cast<in6_addr*>(dst);
+    memcpy(addr->s6_addr, binaddr.ipv6_addr, sizeof(addr->s6_addr));
+  } else {
+    return -1;
+  }
+  return 0;
+}
+
+namespace net {
+
+size_t getBinAddr(void* dest, const std::string& ip)
+{
+  size_t len = 0;
+  addrinfo* res;
+  if(callGetaddrinfo(&res, ip.c_str(), 0, AF_UNSPEC,
+                     0, AI_NUMERICHOST, 0) != 0) {
+    return len;
+  }
+  WSAAPI_AUTO_DELETE<addrinfo*> resDeleter(res, freeaddrinfo);
+  for(addrinfo* rp = res; rp; rp = rp->ai_next) {
+    sockaddr_union su;
+    memcpy(&su, rp->ai_addr, rp->ai_addrlen);
+    if(rp->ai_family == AF_INET) {
+      len = sizeof(in_addr);
+      memcpy(dest, &(su.in.sin_addr), len);
+      break;
+    } else if(rp->ai_family == AF_INET6) {
+      len = sizeof(in6_addr);
+      memcpy(dest, &(su.in6.sin6_addr), len);
+      break;
+    }
+  }
+  return len;
+}
+
+bool verifyHostname(const std::string& hostname,
+                    const std::vector<std::string>& dnsNames,
+                    const std::vector<std::string>& ipAddrs,
+                    const std::string& commonName)
+{
+  if(util::isNumericHost(hostname)) {
+    if(ipAddrs.empty()) {
+      return commonName == hostname;
+    }
+    // We need max 16 bytes to store IPv6 address.
+    unsigned char binAddr[16];
+    size_t addrLen = getBinAddr(binAddr, hostname);
+    if(addrLen == 0) {
+      return false;
+    }
+    for(std::vector<std::string>::const_iterator i = ipAddrs.begin(),
+          eoi = ipAddrs.end(); i != eoi; ++i) {
+      if(addrLen == (*i).size() &&
+         memcmp(binAddr, (*i).c_str(), addrLen) == 0) {
+        return true;
+      }
+    }
+  } else {
+    if(dnsNames.empty()) {
+      return util::tlsHostnameMatch(commonName, hostname);
+    }
+    for(std::vector<std::string>::const_iterator i = dnsNames.begin(),
+          eoi = dnsNames.end(); i != eoi; ++i) {
+      if(util::tlsHostnameMatch(*i, hostname)) {
+        return true;
+      }
+    }
+  }
+  return false;
+}
+
+} // namespace net
+
 } // namespace aria2
