@@ -25,7 +25,7 @@ class RequestTest:public CppUnit::TestFixture {
   CPPUNIT_TEST(testInnerLinkInReferer);
   CPPUNIT_TEST(testGetURIHost);
   CPPUNIT_TEST_SUITE_END();
-  
+
 public:
   void testSetUri1();
   void testSetUri2();
@@ -97,9 +97,14 @@ void RequestTest::testRedirectUri()
   Request req;
   req.supportsPersistentConnection(false);
   req.setUri("http://aria.rednoah.com:8080/aria2/index.html");
-  
-  bool v2 = req.redirectUri("http://aria.rednoah.co.jp/");
-  CPPUNIT_ASSERT(v2);
+
+  // See port number is preserved.
+  CPPUNIT_ASSERT(req.redirectUri("/foo"));
+  CPPUNIT_ASSERT_EQUAL(std::string("http://aria.rednoah.com:8080/foo"),
+                       req.getCurrentUri());
+  CPPUNIT_ASSERT_EQUAL(1, req.getRedirectCount());
+
+  CPPUNIT_ASSERT(req.redirectUri("http://aria.rednoah.co.jp/"));
   // persistent connection flag is set to be true after redirection
   CPPUNIT_ASSERT(req.supportsPersistentConnection());
   // uri must be the same
@@ -118,20 +123,31 @@ void RequestTest::testRedirectUri()
   CPPUNIT_ASSERT_EQUAL(std::string(""), req.getFile());
   CPPUNIT_ASSERT_EQUAL(std::string(""), req.getQuery());
   // See redirect count is incremented.
-  CPPUNIT_ASSERT_EQUAL((unsigned int)1, req.getRedirectCount());
+  CPPUNIT_ASSERT_EQUAL(2, req.getRedirectCount());
 
   // Give abosulute path
   CPPUNIT_ASSERT(req.redirectUri("/abspath/to/file"));
   CPPUNIT_ASSERT_EQUAL(std::string("http://aria.rednoah.co.jp/abspath/to/file"),
                        req.getCurrentUri());
-  CPPUNIT_ASSERT_EQUAL((unsigned int)2, req.getRedirectCount());
+  CPPUNIT_ASSERT_EQUAL(3, req.getRedirectCount());
 
   // Give relative path
   CPPUNIT_ASSERT(req.redirectUri("relativepath/to/file"));
   CPPUNIT_ASSERT_EQUAL(std::string("http://aria.rednoah.co.jp/abspath/to/"
                                    "relativepath/to/file"),
                        req.getCurrentUri());
-  CPPUNIT_ASSERT_EQUAL((unsigned int)3, req.getRedirectCount());
+  CPPUNIT_ASSERT_EQUAL(4, req.getRedirectCount());
+
+  // Give network-path reference
+  CPPUNIT_ASSERT(req.redirectUri("//host/to/file"));
+  CPPUNIT_ASSERT_EQUAL(std::string("http://host/to/file"), req.getCurrentUri());
+  CPPUNIT_ASSERT_EQUAL(5, req.getRedirectCount());
+
+  // http:// in query part
+  CPPUNIT_ASSERT(req.redirectUri("/abspath?uri=http://foo"));
+  CPPUNIT_ASSERT_EQUAL(std::string("http://host/abspath?uri=http://foo"),
+                       req.getCurrentUri());
+  CPPUNIT_ASSERT_EQUAL(6, req.getRedirectCount());
 }
 
 void RequestTest::testRedirectUri2()
@@ -158,7 +174,7 @@ void RequestTest::testRedirectUri2()
   CPPUNIT_ASSERT_EQUAL(std::string("http://aria.rednoah.com/"),
                        req.getPreviousUri());
 }
-  
+
 void RequestTest::testResetUri()
 {
   Request req;
@@ -182,7 +198,7 @@ void RequestTest::testResetUri()
   CPPUNIT_ASSERT_EQUAL(std::string("http"), req.getProtocol());
   CPPUNIT_ASSERT_EQUAL((uint16_t)8080, req.getPort());
   CPPUNIT_ASSERT_EQUAL(std::string("aria.rednoah.com"), req.getHost());
-  CPPUNIT_ASSERT_EQUAL(std::string("/aria2"), req.getDir());
+  CPPUNIT_ASSERT_EQUAL(std::string("/aria2/"), req.getDir());
   CPPUNIT_ASSERT_EQUAL(std::string("index.html"), req.getFile());
   CPPUNIT_ASSERT_EQUAL(std::string(""), req.getQuery());
 }
